########################################################################
#
# Copyright (C) 2021,2022
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: aips2-request@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''This contains functions to inject the ``casaguijs`` library into the
generated HTML that is used to display the Bokeh plots that ``casagui``'s
applications produce'''
from os.path import dirname, join, basename
from bokeh import resources
from ...utils import path_to_url, static_vars, have_network
from bokeh import __version__ as bokeh_version

CASALIB_VERSION = "0.0.2"
CASAGUIJS_VERSION = "0.0.6"

@static_vars(initialized=False,do_local_subst=not have_network( ))
def initialize_bokeh( js=None, bokeh=None, bokeh_dev=0 ):
    """Initialize `bokeh` for use with the ``casaguijs`` extensions.

    The ``casaguijs`` extensions for Bokeh are built into a stand-alone
    external JavaScript file constructed for the specific version of
    Bokeh that is being used. The URL or the local path to a locally
    built verion must be supplied as the ``bokeh`` parameter. If ``bokeh``
    is ``None`` then the default is to use the published verison of
    ``casaguijs`` extensions built and installed for `HTTP` access.

    Parameters
    ----------
    js: None or str or list of str
        Bokeh independent javascript library which loaded before any
        Bokeh libraries. This path could be a local path, a URL or None.
        None is the default in which case it loads the published library
        for the current version of ``casaguijs`` and ``bokeh``
    bokeh: None or str or list of str
        Bokeh dependent javascript library which is loaded after all
        other Bokeh libraries have been loaded. This path could be a
        local path, a URL or None. None is the default in which case
        it loads the published library for the current version of
        ``casaguijs`` and ``bokeh``
    bokeh_dev: int, optional
        Chrome caches the javascript files. This parameter allows for
        specifying `bokeh_dev` allows for including a development version
        for incremental updates to the ``bokeh`` library.

    Example
    -------
    from casagui.bokeh.state import initialize_bokeh
    initialize_bokeh( bokeh="/tmp/casagui/kernels/python/casaguijs/dist/casaguijs.min.js" )
    """

    if initialize_bokeh.initialized:
        ### only initialize once...
        return

    ###
    ### if no network is available, substitute local JavaScript libraries
    ### for remote URLs
    ###

    library_hashes = {
        ### --------------------------------------------------------------------------------------------------
        ### Generate hashes with:
        ### echo `cat casaguijs-v0.0.2.0-b2.3.js | openssl dgst -sha384 -binary | openssl base64 -A`
        ### ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---  ---
        ### NOTE: with the conversion from Bokeh 2.3.3 to 2.4.1, it seems that the security hashes are no
        ### longer included in the generated HTML.
        ### --------------------------------------------------------------------------------------------------
        'casaguijs-v0.0.2.0-b2.3.js': 'EtB9H3ooIK3inPGx9RsRXeUv/COHtArEjhZUr7P75GBUPl+lAoGH/tqoAc1sV5jD',
        'casaguijs-v0.0.2.0-b2.3.min.js': 'jAFFXRC9B93jSanS1dmqo+l/3rkcBqM48uTMN+SFcY7GPC/IPeEoI+p+Za5ztkcm',
        'casaguijs-v0.0.3.0-b2.4.min.js': '23DOS7ISXIG8BRyIsD7vINUllDbo8NDCozIGBy7jGC+M14Z+axfF7j4bruA2ZnzL',
        'casaguijs-v0.0.3.0-b2.4.js': 'OmIWKy37YXZf973kemdoFd2L/h0I9AY/hsP5PMdk5/g5xlgj16yk6b0cgO01UMxM'
    }
    ### --------------------------------------------------------------------------------------------------
    ### casagui will ship with local versions of the Bokeh and casaguijs libraries which can be used
    ### when there is no network connectivity...
    ### --------------------------------------------------------------------------------------------------
    local_libraries = {
        'bokeh-2.4.1.min.js': join( dirname(__file__), 'js', 'bokeh-2.4.1.min.js' ),
        'bokeh-gl-2.4.1.min.js':  join( dirname(__file__), 'js', 'bokeh-gl-2.4.1.min.js' ),
        'bokeh-widgets-2.4.1.min.js':  join( dirname(__file__), 'js', 'bokeh-widgets-2.4.1.min.js' ),
    }

    bokeh_major_minor = '.'.join(bokeh_version.split('.')[0:2])

    bokehlib = None
    bokehlib_libs = None
    bokehlib_url = None

    ###
    ### Initialize the Bokeh dependent library(ies)
    ###
    if bokeh is None:
        bokehlib = f"casaguijs-v{CASAGUIJS_VERSION}.{bokeh_dev}-b{bokeh_major_minor}.min.js"
        if initialize_bokeh.do_local_subst:
            bokehlib_url = path_to_url( join( dirname(__file__), 'js', bokehlib ) )
        else:
            ### ------------------------------------------------------------------------------------------
            ### should potentially find a better download location...
            ### ------------------------------------------------------------------------------------------
            ### poor availablity (fails to load, server not available etc (even from NRAO network):
            #bokehlib_url = f"https://casa.nrao.edu/download/javascript/casaguijs/{CASAGUIJS_VERSION}/{bokehlib}"
            ### does not work at all (CERT issue?):
            #bokehlib_url = f"https://raw.githubusercontent.com/casangi/casagui-js/main/runtime/{CASAGUIJS_VERSION}/{bokehlib}"
            bokehlib_url = f"https://cdn.jsdelivr.net/gh/casangi/casagui-js@main/runtime/{CASAGUIJS_VERSION}/{bokehlib}"

        bokehlib_libs = [ bokehlib_url ]
    else:
        bokehlib_libs = [ bokeh ] if isinstance(bokeh,str) else bokeh
        bokehlib_libs = list(map( path_to_url, bokehlib_libs ))

    jslib = None
    jslib_libs = None
    jslib_url = None

    ###
    ### Initialize the Bokeh independent, javascript library(ies)
    ###
    if js is None:
        jslib = f"casalib-v{CASALIB_VERSION}.min.js"
        if initialize_bokeh.do_local_subst:
            jslib_url = path_to_url( join( dirname(__file__), 'js', jslib ) )
        else:
            ### ------------------------------------------------------------------------------------------
            ### should potentially find a better download location...
            ### ------------------------------------------------------------------------------------------
            ### poor availablity (fails to load, server not available etc (even from NRAO network):
            #jslib_url = f"https://casa.nrao.edu/download/javascript/casaguijs/{CASAGUIJS_VERSION}/{jslib}"
            ### does not work at all (CERT issue?):
            #jslib_url = f"https://raw.githubusercontent.com/casangi/casagui-js/main/runtime/{CASAGUIJS_VERSION}/{jslib}"
            jslib_url = f"https://cdn.jsdelivr.net/gh/casangi/casagui-js@main/runtime/{CASAGUIJS_VERSION}/{jslib}"
        jslib_libs = [ jslib_url ]
    else:
        jslib_libs = [ js ] if isinstance(js,str) else js
        jslib_libs = list(map( path_to_url, jslib_libs ))

    ###
    ### substitute our function for the Bokeh function that retrieves
    ### the security hashes for the javascript files...
    ###
    resources.JSResources._old_hashes = resources.JSResources.hashes     # pylint: disable=protected-access
                                                                         # could find no other way to add our library
                                                                         # to the libraries loaded by Bokeh; maintainers
                                                                         # had no solutions to offer, see:
                                                                         # https://discourse.bokeh.org/t/pre-built-extensions-without-bokeh-server/7992
    def hashes( self ):
        result = self._old_hashes                                        # pylint: disable=protected-access
        if bokehlib is not None and bokehlib in library_hashes:
            result[bokehlib_url] = library_hashes[bokehlib]
        if jslib is not None and jslib in library_hashes:
            result[jslib_url] = library_hashes[jslib]
        return result

    ###
    ### substitute our function for the Bokeh function that retrieves
    ### the javascript files...
    ###
    resources.JSResources._old_js_files = resources.JSResources.js_files # pylint: disable=protected-access
    def js_files( self ):
        if initialize_bokeh.do_local_subst:
            result = [ ]
            for url in self._old_js_files:                               # pylint: disable=protected-access
                lib = basename(url)
                if lib in local_libraries:
                    result.append( path_to_url(local_libraries[lib]) )
                else:
                    result.append( url )
        else:
            result = self._old_js_files                                  # pylint: disable=protected-access
        result = jslib_libs + result + bokehlib_libs
        return result

    resources.JSResources.hashes = property(hashes)
    resources.JSResources.js_files = property(js_files)
    initialize_bokeh.initialized = True



import os
import asyncio
import urllib
import tarfile
#from casagui.bokeh.state import initialize_bokeh
#initialize_bokeh( ".../casagui/casaguijs/dist/casaguijs.min.js" )        ### local build

from casagui.apps import MakeMask

##
## demo measurement set to use
##
image_path = 'refim_point_withline.image'
##
## where to fetch the demo measurement set
##
image_url = "https://casa.nrao.edu/download/devel/casavis/data/refim_point_withline-image.tar.gz"

if not os.path.isdir(image_path):
    tstream = urllib.request.urlopen(image_url)
    tar = tarfile.open(fileobj=tstream, mode="r:gz")
    tar.extractall()

print(repr(MakeMask( image_path )( )))



from bokeh.core import properties
from bokeh.models.widgets import AbstractIcon

class SVGIcon(AbstractIcon):
    """
    The SVGIcon can be used to add SVG based icons to buttons, menus etc.

    See https://github.com/holoviz/panel/issues/1586 for motivation, possibilities 
    and requirements.

    Args:
        AbstractIcon (Abstract Class): Abstract class that can be used to 
        build custom Icon classes from.
    """
    

    __implementation__ = "svg_icon.ts"

    icon_name = properties.String()
    size = properties.Float()
    fill_color = properties.String()
    spin_duration = properties.Int()



########################################################################
#
# Copyright (C) 2022
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: aips2-request@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''implementation of the ``MakeMask`` application for interactive creation
of channel masks'''
import asyncio
from bokeh.layouts import row, column
from bokeh.plotting import show
from bokeh.models import Button, CustomJS
from casagui.toolbox import CubeMask
from casagui.bokeh.components import SVGIcon

class MakeMask:
    '''Class that can be used to launch a makemask GUI with ``MakeMask('test.image')( )``.'''

    def __init__( self, image ):
        '''create a ``makemask`` object which will display image planes from a CASA
        image and allow the user to draw masks for each channel.

        Parameters
        ----------
        image: str
            path to CASA image for which interactive masks will be drawn
        '''
        self._cube = CubeMask(image)
        self._done = None
        self._help = None
        self._help_text = None
        self._layout = None

    def _launch_gui( self ):
        '''create and show GUI
        '''
        width = 35
        height = 35
        self._done = Button( label="", button_type="danger", max_width=width, max_height=height, name='done',
                             icon=SVGIcon(icon_name='makemask-done', size=1.4) )
        self._help = Button( label="", max_width=width, max_height=height, name='help',
                             icon=SVGIcon(icon_name='help', size=1.4) )

        ### CubeMask provides a help table
        self._help_text = self._cube.help( rows=[ '<tr><td><i>red check button</i></td><td>clicking the red check button will close the dialog and return masks to python</td></tr>' ] )

        self._layout = column( self._cube.image( ),
                               row( self._cube.slider( ), self._help, self._done ),
                               self._help_text )

        self._done.js_on_click( CustomJS( args=dict( obj=self._cube.js_obj( ) ),
                                          code='''obj.done( )''' ) )                  ### CubeMask should return javascript object which contains
                                                                                      ### a done function to stop cube masking
        self._help.js_on_click( CustomJS( args=dict( help=self._help_text ),
                                          code='''if ( help.visible == true ) help.visible = false
                                                  else help.visible = true''' ) )
        self._cube.connect( )
        show(self._layout)

    def _asyncio_loop( self ):
        '''return the event loop which can be mixed in with an existing event loop
        to allow GUI and websocket events to be processed.
        '''
        return self._cube.loop( )

    def __call__( self, loop=asyncio.get_event_loop( ) ):
        '''Display GUI using the event loop specified by ``loop``.

        Parameters
        ----------
        loop: event loop object
            defaults to standard asyncio event loop.

        Example:
            Create ``iclean`` object and display::

                print( "Result: %s" %
                       iclean( vis='refim_point_withline.ms', imagename='test', imsize=512,
                               cell='12.0arcsec', specmode='cube',
                               interpolation='nearest', ... )( ) )
        '''
        try:
            loop.run_until_complete(self.show( ))
            loop.run_forever( )
        except KeyboardInterrupt:
            print('\nInterrupt received, stopping GUI...')

        return self.result( )

    def show( self ):
        '''Get the makemask event loop to use for running the ``makemask`` GUI
        as part of an external event loop.
        '''
        self._launch_gui( )
        return self._asyncio_loop( )

    def result( self ):
        '''Retrieve the masks that have been drawn by the user. The return
        value is a dictionary with two elements ``masks`` and ``polys``.

        The value of the ``masks`` element is a dictionary that is indexed by
        tuples of ``(stokes,chan)`` and the value of each element is a list
        whose elements describe the polygons drawn on the channel represented
        by ``(stokes,chan)``. Each polygon description in this list has a
        polygon index (``p``) and a x/y translation (``d``).

        The value of the ``polys`` element is a dictionary that is indexed by
        polygon indexes. The value of each polygon index is a dictionary containing
        ``type`` (whose value is either ``'rect'`` or ``'poly``) and ``geometry``
        (whose value is a dictionary containing ``'xs'`` and ``'ys'`` (which are
        the x and y coordinates that define the polygon).
        '''
        return self._cube.result( )



#-------------------------------------------------------------------------------
# Copyright (c) 2012 - 2021, Anaconda, Inc., and Bokeh Contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# Neither the name of Anaconda nor the names of any contributors
# may be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
#
#-------------------------------------------------------------------------------
#-- With the conversion from Bokeh 2.3.3 to 2.4.1 the transport of numpy      --
#-- arrays was broken. In time we may need to develop our own packing and     --
#-- unpacking routines for python/javascript however for the time being       --
#-- continuing to use transform_array( ) below from 2.3.3 solved the problem. --
#-------------------------------------------------------------------------------
#
#  This file is a subset of bokeh/util/serialization.py from the Bokeh
#  repository:
#
#  bash$ git config --get remote.origin.url
#  https://github.com/bokeh/bokeh.git
#  bash$
#
#--- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#  with the most recent release tag:
#
#  bash$ git show-ref --tags| tail -1
#  8ce5e80640cd16c9afd0045fcec42404ec59a818 refs/tags/2.3.2rc2
#  bash$
#
#--- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#  with these further revisions:
#
#  bash$ git log --pretty=oneline --no-merges 8ce5e80640cd16c9..HEAD bokeh/util/serialization.py
#  af9183f00dcff18a51a6a5af75099668463e3012 Add support for lazy annotations boilerplate (#11220)
#  bash$
#
#--- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#  which includes this difference:
#
#  bash$ git diff 8ce5e80640cd16c9afd0045fcec42404ec59a818 bokeh/util/serialization.py
#  diff --git a/bokeh/util/serialization.py b/bokeh/util/serialization.py
#  index a73a06358..eeeecf29a 100644
#  --- a/bokeh/util/serialization.py
#  +++ b/bokeh/util/serialization.py
#  @@ -18,6 +18,8 @@ performance and efficiency. The list of supported dtypes is:
#   #-----------------------------------------------------------------------------
#   # Boilerplate
#   #-----------------------------------------------------------------------------
#  +from __future__ import annotations
#  +
#   import logging # isort:skip
#   log = logging.getLogger(__name__)
#
#  bash$
#
#-----------------------------------------------------------------------------
'''
Functions for helping with serialization and deserialization of
Bokeh objects.

Certain NumPy array dtypes can be serialized to a binary format for
performance and efficiency. The list of supported dtypes is:

{binary_array_types}

'''

# pylint: skip-file

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
import base64
import datetime as dt
import sys
import uuid
from math import isinf, isnan
from threading import Lock

# External imports
import numpy as np

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

#pd = import_optional('pandas')

BINARY_ARRAY_TYPES = {
    np.dtype(np.float32),
    np.dtype(np.float64),
    np.dtype(np.uint8),
    np.dtype(np.int8),
    np.dtype(np.uint16),
    np.dtype(np.int16),
    np.dtype(np.uint32),
    np.dtype(np.int32),
}

DATETIME_TYPES = {
    dt.time,
    dt.datetime,
    np.datetime64,
}

NP_EPOCH = np.datetime64(0, 'ms')
NP_MS_DELTA = np.timedelta64(1, 'ms')

DT_EPOCH = dt.datetime.utcfromtimestamp(0)

#__doc__ = format_docstring(__doc__, binary_array_types="\n".join("* ``np." + str(x) + "``" for x in BINARY_ARRAY_TYPES))

__all__ = (
    'transform_array',
    'convert_datetime_array',
    'serialize_array',
    'encode_base64_dict',
    'encode_binary_dict',
    'decode_base64_dict',
    'transform_array_to_list'
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

def convert_datetime_array(array):
    ''' Convert NumPy datetime arrays to arrays to milliseconds since epoch.

    Args:
        array : (obj)
            A NumPy array of datetime to convert

            If the value passed in is not a NumPy array, it will be returned as-is.

    Returns:
        array

    '''

    if not isinstance(array, np.ndarray):
        return array

    # not quite correct, truncates to ms..
    if array.dtype.kind == 'M':
        array =  array.astype('datetime64[us]').astype('int64') / 1000.

    elif array.dtype.kind == 'm':
        array = array.astype('timedelta64[us]').astype('int64') / 1000.

    # (bev) special case dates, not great
    elif array.dtype.kind == 'O' and len(array) > 0 and isinstance(array[0], dt.date):
        try:
            array = array.astype('datetime64[us]').astype('int64') / 1000.
        except Exception:
            pass

    return array

def array_encoding_disabled(array):
    ''' Determine whether an array may be binary encoded.

    The NumPy array dtypes that can be encoded are:

    {binary_array_types}

    Args:
        array (np.ndarray) : the array to check

    Returns:
        bool

    '''

    # disable binary encoding for non-supported dtypes
    return array.dtype not in BINARY_ARRAY_TYPES

def transform_array(array, force_list=False, buffers=None):
    ''' Transform a NumPy arrays into serialized format

    Converts un-serializable dtypes and returns JSON serializable
    format

    Args:
        array (np.ndarray) : a NumPy array to be transformed
        force_list (bool, optional) : whether to only output to standard lists
            This function can encode some dtypes using a binary encoding, but
            setting this argument to True will override that and cause only
            standard Python lists to be emitted. (default: False)

        buffers (set, optional) :
            If binary buffers are desired, the buffers parameter may be
            provided, and any columns that may be sent as binary buffers
            will be added to the set. If None, then only base64 encoding
            will be used (default: None)

            If force_list is True, then this value will be ignored, and
            no buffers will be generated.

            **This is an "out" parameter**. The values it contains will be
            modified in-place.


    Returns:
        JSON

    '''

    array = convert_datetime_array(array)

    return serialize_array(array, force_list=force_list, buffers=buffers)

def transform_array_to_list(array):
    ''' Transforms a NumPy array into a list of values

    Args:
        array (np.nadarray) : the NumPy array series to transform

    Returns:
        list or dict

    '''
    if (array.dtype.kind in ('u', 'i', 'f') and (~np.isfinite(array)).any()):
        transformed = array.astype('object')
        transformed[np.isnan(array)] = 'NaN'
        transformed[np.isposinf(array)] = 'Infinity'
        transformed[np.isneginf(array)] = '-Infinity'
        return transformed.tolist()
    #if (array.dtype.kind == 'O' and pd and pd.isnull(array).any()):
    #    transformed = array.astype('object')
    #    transformed[pd.isnull(array)] = 'NaN'
    #    return transformed.tolist()
    return array.tolist()

def serialize_array(array, force_list=False, buffers=None):
    ''' Transforms a NumPy array into serialized form.

    Args:
        array (np.ndarray) : the NumPy array to transform
        force_list (bool, optional) : whether to only output to standard lists
            This function can encode some dtypes using a binary encoding, but
            setting this argument to True will override that and cause only
            standard Python lists to be emitted. (default: False)

        buffers (set, optional) :
            If binary buffers are desired, the buffers parameter may be
            provided, and any columns that may be sent as binary buffers
            will be added to the set. If None, then only base64 encoding
            will be used (default: None)

            If force_list is True, then this value will be ignored, and
            no buffers will be generated.

            **This is an "out" parameter**. The values it contains will be
            modified in-place.

    Returns:
        list or dict

    '''
    if isinstance(array, np.ma.MaskedArray):
        array = array.filled(np.nan)  # Set masked values to nan
    if (array_encoding_disabled(array) or force_list):
        return transform_array_to_list(array)
    if not array.flags['C_CONTIGUOUS']:
        array = np.ascontiguousarray(array)
    if buffers is None:
        return encode_base64_dict(array)
    return encode_binary_dict(array, buffers)

def encode_binary_dict(array, buffers):
    ''' Send a numpy array as an unencoded binary buffer

    The encoded format is a dict with the following structure:

    .. code:: python

        {
            '__buffer__' :  << an ID to locate the buffer >>,
            'shape'      : << array shape >>,
            'dtype'      : << dtype name >>,
            'order'      : << byte order at origin (little or big)>>
        }

    Args:
        array (np.ndarray) : an array to encode

        buffers (set) :
            Set to add buffers to

            **This is an "out" parameter**. The values it contains will be
            modified in-place.

    Returns:
        dict

    '''
    buffer_id = make_id()
    buf = (dict(id=buffer_id), array.tobytes())
    buffers.append(buf)

    return {
        '__buffer__'  : buffer_id,
        'shape'       : array.shape,
        'dtype'       : array.dtype.name,
        'order'       : sys.byteorder,
    }

def encode_base64_dict(array):
    ''' Encode a NumPy array using base64:

    The encoded format is a dict with the following structure:

    .. code:: python

        {
            '__ndarray__' : << base64 encoded array data >>,
            'shape'       : << array shape >>,
            'dtype'       : << dtype name >>,
        }

    Args:

        array (np.ndarray) : an array to encode

    Returns:
        dict

    '''
    return {
        '__ndarray__' : base64.b64encode(array.data).decode('utf-8'),
        'shape'       : array.shape,
        'dtype'       : array.dtype.name,
        'order'       : sys.byteorder,
    }

def decode_base64_dict(data):
    ''' Decode a base64 encoded array into a NumPy array.

    Args:
        data (dict) : encoded array data to decode

    Data should have the format encoded by :func:`encode_base64_dict`.

    Returns:
        np.ndarray

    '''
    b64 = base64.b64decode(data['__ndarray__'])
    array = np.copy(np.frombuffer(b64, dtype=data['dtype']))
    if len(data['shape']) > 1:
        array = array.reshape(data['shape'])
    return array



########################################################################
#
# Copyright (C) 2022
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: aips2-request@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''implementation of the ``InteractiveClean`` application for interactive control
of tclean'''
import os
import copy
import asyncio
import shutil
import websockets
from uuid import uuid4
from bokeh.models import Button, TextInput, Div, Range1d, LinearAxis, CustomJS, Spacer
from bokeh.plotting import ColumnDataSource, figure, show
from bokeh.layouts import column, row, Spacer
from ..utils import resource_manager

try:
    ## gclean version number needed for proper interactive clean behavior
    from casatasks.private.imagerhelpers._gclean import _GCV001
    from casatasks.private.imagerhelpers._gclean import gclean as _gclean
except:
    try:
        ###
        ### enable this warning when casa6 a usable _gclean.py (i.e. compatibility is not the default)
        ###
        #print('warning: using tclean compatibility layer...')
        from ..private.compatibility.casatasks.private.imagerhelpers._gclean import _GCV001
        from ..private.compatibility.casatasks.private.imagerhelpers._gclean import gclean as _gclean
    except:
        _gclean = None
        from casagui.utils import warn_import
        warn_import('casatasks')

from casagui.utils import find_ws_address, convert_masks
from casagui.toolbox import CubeMask
from casagui.bokeh.components import SVGIcon
from casagui.bokeh.sources import DataPipe
from ..utils import DocEnum

class MaskMode(DocEnum):
    '''Different masking modes available in addition to a user supplied mask'''
    PB = 1, 'primary beam mask'
    AUTOMT = 2, 'multi-threshold auto masking'

class InteractiveClean:
    '''InteractiveClean(...) implements interactive clean using Bokeh

    This class allows for creation of an Bokeh based GUI for iterative, interactive execution of
    ``tclean``. It allows for drawing the mask that will be used by ``tclean`` as well as running
    and stoping clean cycles.

    Example:
        First the interactive clean GUI is created with::

            ic = InteractiveClean( vis='refim_point_withline.ms', imagename='test', imsize=512,
                                   cell='12.0arcsec', specmode='cube', interpolation='nearest', ... )

        and then the GUI can be started with::

            ic( )

    The GUI will be displayed in a web browser and control will return to Python when the
    user clicks the ``STOP`` button on the GUI. When the ``STOP`` button is colored *red*,
    it means that clicking it will close the tab and return control to Python. When it is
    colored *orange* it means that clicking it will stop ``tclean`` at the next stopping
    point.

    Masks can be specified either with a lasso selection tool or with a box selection tool
    in the _image box_. As the cursor moves around the _image box_, the spectrum through the
    image cube beneath the cursor location is displayed in the _spectrum box_. Quickly tapping
    _mouse one_ within the _image box_ will freeze updates of the _spectrum box_ allowing
    a spectrum of interest to be selected.

    Quickly tapping _mouse one_ within the _spectrum box_ cause the image display to update
    to the channel indicated by the vertical line which track with the cursor in the
    _spectrum box_.

    The _convergence box_ plots the _total flux_ and _peak residual_ to give an indication
    of the convergence quailty. A _statistics table_ provides some key image statistics. All
    of these boxes are updated as the clean operation progresses.

    The cleaning interval is determined by the text inputs along the top of the GUI which
    allow the user to set the ``nmajor``, ``niter``, ``cycleiter``, ``cyclefactor``, and ``threshold``.
    The three buttons ( _run one interval_, _run until stoping criteral is reached_, and
    _stop_) allow the user to control the clean run.

    At the bottom of the GUI, footnotes accumulate which includes the Python calls to
    ``tclean`` that have been made.

    *Please see tclean documentation for details about the arguments summarized here.*

    Args:
        vis (str, :obj:`list` of :obj:`str`): Name of input visibility file(s). A single path may
            be specified, e.g. "ngc5921.ms", or a list of paths may be specified, e.g.
            ``['ngc5921a.ms','ngc5921b.ms']``.
        imagename (str): Name of the output images (without a suffix).
        mask (str or MaskMode): user specified CASA imaging mask cube (path) or MaskMode enum indicating
            mask to generate
        imsize (int, :obj:`list` of int): Number of pixels in a image plane, for example ``[350,250]``.
            ``500`` implies ``[500,500]``. The number of pixels must be even and factorizable by 2,3,5,7 only
            due to FFT limitations.
        cell (str, :obj:`list` of str): Cell size, e.g. ``['0.5arcsec,'0.5arcsec']``. ``'1arcsec'`` is
            equivalent to ``['1arcsec','1arcsec']``
        specmode (str): Spectral definition mode (mfs,cube,cubedata, cubesource)
            * mode='mfs' : Continuum imaging with only one output image channel
            * mode='cube' : Spectral line imaging with one or more channels
            * mode='cubedata' : Spectral line imaging with one or more channels
        nchan (int): Number of channels in the output image. For default (=-1),
            the number of channels will be automatically determined based on data
            selected by 'spw' with 'start' and 'width'.
        start (int, str): First channel (e.g. start=3,start='1.1GHz',start='15343km/s'.
        width (int, str): Channel width (e.g. width=2,width='0.1MHz',width='10km/s') of output cube images
            specified by data channel number (integer), velocity (string with a unit), or frequency (string
            with a unit).
        interpolation (str): Spectral interpolation (nearest,linear,cubic). Interpolation
            rules to use when binning data channels onto image channels and evaluating
            visibility values at the centers of image channels.
        gridder (str): Gridding options (standard, wproject, widefield, mosaic, awproject)
        pblimit (float): PB gain level at which to cut off normalizations
        deconvolver (str): Minor cycle algorithm (hogbom,clark,multiscale,mtmfs,mem,clarkstokes)
        nmajor (int): Maximum number of major cycle iterations
        niter (int): Maximum total number of iterations
        threshold (str | float): Stopping threshold (number in units of Jy, or string).
        cycleniter (int): Maximum number of minor-cycle iterations (per plane) before triggering a major cycle
        cyclefactor (float): Scaling on PSF sidelobe level to compute the minor-cycle stopping threshold.
        scales (:obj:`list` of int): List of scale sizes (in pixels) for multi-scale algorithms

    '''
    def __stop( self ):
        resource_manager.stop_asyncio_loop()
        if self._control_server is not None and self._control_server.ws_server.is_serving( ):
            resource_manager.stop_asyncio_loop()
        if self._converge_server is not None and self._converge_server.ws_server.is_serving( ):
            resource_manager.stop_asyncio_loop()

    def _abort_handler( self, loop, err ):
        self._error_result = err
        self.__stop( )

    def __init__( self, vis, imagename, mask=None, field='', spw='', timerange='', uvrange='', antenna='', scan='', intent='', datacolumn='corrected',
                  nterms=int(2), imsize=[100], cell=[ ], phasecenter='', stokes='I', specmode='cube', reffreq='', nchan=-1, start='', width='',
                  outframe='LSRK', interpolation='linear', perchanweightdensity=True, gridder='standard', wprojplanes=int(1), mosweight=True,
                  psterm=False, wbawp=True, usepointing=False, conjbeams=False, pointingoffsetsigdev=[  ], pblimit=0.2, deconvolver='hogbom',
                  niter=0, threshold='0.1Jy', nsigma=0.0, cycleniter=-1, cyclefactor=1.0, scales=[], restoringbeam='', pbcor=False,
                  weighting='natural', robust=float(0.5), npixels=0, gain=float(0.1), sidelobethreshold=3.0, noisethreshold=5.0,
                  lownoisethreshold=1.5, negativethreshold=0.0, minbeamfrac=0.3, growiterations=75, dogrowprune=True, minpercentchange=-1.0,
                  fastnoise=True, savemodel='none', parallel=False, nmajor=1 ):

        if deconvolver == 'mtmfs':
            raise RuntimeError("deconvolver task does not support 'mtmf' deconvolver")

        ###
        ### used by data pipe (websocket) initialization function
        ###
        self.__pipes_initialized = False
        self._mask_history = [ ]

        ###
        ### color specs
        ###
        self._color = { 'residual': 'black',
                        'flux':     'forestgreen' }

        ###
        ### Auto Masking et al.
        ###     if the user has specified a mask cube, it OVERRIDES the default, generated mask name
        self._mask_path = ''
        self._usemask = 'user'
        if isinstance( mask, MaskMode ):
            if mask == MaskMode.AUTOMT:
                self._usemask = 'auto-multithresh'
            elif mask == MaskMode.PB:
                self._usemask = 'pb'
            else:
                raise RuntimeError( 'internal consistence error for MaskMode' )
        elif isinstance( mask, str ) and os.path.isdir( mask ):
            ### override default mask name
            self._mask_path = mask

        ###
        ### clean generator
        ###
        if _gclean is None:
            raise RuntimeError('casatasks gclean interface is not available')

        self._clean = _gclean( vis=vis, imagename=imagename, field=field, spw=spw, timerange=timerange,  uvrange=uvrange, antenna=antenna, scan=scan,
                               intent=intent, datacolumn=datacolumn, nterms=nterms, imsize=imsize, cell=cell, phasecenter=phasecenter, stokes=stokes,
                               specmode=specmode, reffreq=reffreq, nchan=nchan, start=start, width=width, outframe=outframe, interpolation=interpolation,
                               perchanweightdensity=perchanweightdensity, gridder=gridder, wprojplanes=wprojplanes, mosweight=mosweight, psterm=psterm,
                               wbawp=wbawp, usepointing=usepointing, conjbeams=conjbeams, pointingoffsetsigdev=pointingoffsetsigdev, pblimit=pblimit,
                               deconvolver=deconvolver, niter=niter, threshold=threshold, nsigma=nsigma, cycleniter=cycleniter, cyclefactor=cyclefactor,
                               scales=scales, restoringbeam=restoringbeam, pbcor=pbcor, weighting=weighting, robust=robust, npixels=npixels, gain=gain,
                               sidelobethreshold=sidelobethreshold, noisethreshold=noisethreshold, lownoisethreshold=lownoisethreshold,
                               negativethreshold=negativethreshold, minbeamfrac=minbeamfrac, growiterations=growiterations, dogrowprune=dogrowprune,
                               minpercentchange=minpercentchange, fastnoise=fastnoise, savemodel=savemodel, parallel=parallel, nmajor=nmajor,
                               usemask=self._usemask, mask=self._mask_path
                      )
        ###
        ### self._convergence_data: accumulated, pre-channel convergence information
        ###                         used by ColumnDataSource
        ###
        self._status = { }
        stopcode, self._convergence_data = next(self._clean)
        if stopcode > 1 and stopcode < 9: # 1: iteration limit hit, 9: major cycle limit hit
            self._clean.finalize()
        if len(self._convergence_data.keys()) == 0:
            raise RuntimeError("No convergence data for iclean. Did tclean exit without any minor cycles?")
        self._convergence_id = str(uuid4( ))
        #print(f'convergence:',self._convergence_id)

        self._status['log'] = Div( text='''<hr style="width:790px">%s''' % ''.join([ f'<p style="width:790px">{cmd}</p>' for cmd in self._clean.cmds( )[-2:] ]) )
        ###                        >>>--------tclean+deconvolve----------------------------------------------------------------------------------------^^^^^
        self._status['stopcode'] = Div( text="<div>initial image</div>" )

        ###
        ### Initial Conditions
        ###
        self._params = { }
        self._params['nmajor'] = nmajor
        self._params['niter'] = niter
        self._params['cycleniter'] = cycleniter
        self._params['threshold'] = threshold
        self._params['cyclefactor'] = cyclefactor
        ###
        ### Polarity plane
        ###
        self._stokes = 0

        ###
        ### GUI Elements
        self._imagename = imagename
        self._residual_path = ("%s.residual" % imagename) if self._clean.has_next() else (self._clean.finalize()['image'])
        if not os.path.isdir(self._mask_path):
            self._mask_path = ("%s.mask" % imagename) if self._clean.has_next() else None
        self._pipe = { 'control': None, 'converge': None }
        self._control = { }
        self._cb = { }
        self._ids = { }
        self._last_mask_breadcrumbs = ''

        ###
        ### The tclean convergence data is automatically generated by tclean and it
        ### accumulates in this object. If the data becomes bigger than these
        ### thresholds, the implementation switches to fetching threshold data
        ### from python for each channel selected in the GUI within the browser.
        ###
        self._threshold_chan = 400
        self._threshold_iterations = 2000

        self._js = { ### initialize state
                     ### --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
                     ### -- convergence_src is used storing state (initialized and convergence data cache below   --
                     ### --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
                     'initialize':      '''if ( ! convergence_src._initialized ) {
                                               convergence_src._initialized = true
                                               window.addEventListener( 'beforeunload',
                                                                        function (e) {
                                                                            ctrl_pipe.send( ids['stop'],
                                                                                            { action: 'stop', value: { } },
                                                                                              undefined ) }
                                                                      )
                                           }''',

                     ### --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
                     ### -- convergence_src._convergence_data is used to store the complete                       --
                     ### --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
                     'update-converge': '''function update_convergence( msg ) {
                                               if ( typeof msg === 'undefined' && '_convergence_data' in convergence_src ) {
                                                   // use complete convergence cache attached to convergence_src...
                                                   // get the convergence data for channel and stokes
                                                   const pos = img_src.cur_chan
                                                   convergence_src.data = convergence_src._convergence_data[pos[1]][pos[0]]
                                                   //                             chan----------------------^^^^^^  ^^^^^^----stokes
                                               } else if ( 'result' in msg ) {
                                                   // update based on msg received from conv_pipe
                                                   convergence_src.data = msg.result.converge
                                               }
                                               convergence_fig.extra_y_ranges['modelFlux'].end = 1.5*Math.max(...convergence_src.data['modelFlux'])
                                               convergence_fig.extra_y_ranges['modelFlux'].start = 0.5*Math.min(...convergence_src.data['modelFlux'])
                                           }''',

                     'clean-refresh':   '''function refresh( clean_msg ) {
                                               let stokes = 0    // later we will receive the polarity
                                                                 // from some widget mechanism...
                                               //img_src.refresh( msg => { if ( 'stats' in msg ) { //  -- this should happen within CubeMask
                                               //                              stat_src.data = msg.stats
                                               //                          }
                                               //                        } )
                                               if ( clean_msg !== undefined && 'iterdone' in clean_msg ) {
                                                 const remaining = parseInt(niter.value) - parseInt(clean_msg['iterdone'])
                                                 niter.value = '' + (remaining < 0 ? 0 : remaining)
                                               }
                                               img_src.refresh( (data) => { if ( 'stats' in data ) cube_obj.update_statistics( data.stats ) } )

                                               if ( clean_msg !== undefined && 'convergence' in clean_msg ) {
                                                   // save convergence information and update convergence using saved state
                                                   if ( clean_msg.convergence === null ) {
                                                       delete convergence_src._convergence_data
                                                       const pos = img_src.cur_chan
                                                       // fetch convergence information for the current channel (pos[1])
                                                       // ...convergence update expects [ stokes, chan ]
                                                       conv_pipe.send( convergence_id, { action: 'update', value: pos }, update_convergence )
                                                   } else {
                                                       convergence_src._convergence_data = clean_msg.convergence
                                                       update_convergence( )
                                                   }
                                               } else {
                                                   const pos = img_src.cur_chan
                                                   // fetch convergence information for the current channel (pos[1])
                                                   conv_pipe.send( convergence_id, { action: 'update', value: pos[1] }, update_convergence )
                                               }
                                           }''',

                       'clean-disable': '''// enabling/disabling tools in self._fig['image'].toolbar.tools does not seem to not work
                                           // self._fig['image'].toolbar.tools.tool_name (e.g. "Box Select", "Lasso Select")
                                           function disable( with_stop ) {
                                               img_src.disable_masking( )
                                               nmajor.disabled = true
                                               niter.disabled = true
                                               cycleniter.disabled = true
                                               threshold.disabled = true
                                               cyclefactor.disabled = true
                                               btns['continue'].disabled = true
                                               btns['finish'].disabled = true
                                               slider.disabled = true
                                               image_fig.disabled = true
                                               spectra_fig.disabled = true
                                               if ( with_stop ) {
                                                   btns['stop'].disabled = true
                                               } else {
                                                   btns['stop'].disabled = false
                                               }
                                           }''',

                       'clean-enable':  '''function enable( only_stop ) {
                                               img_src.enable_masking( )
                                               nmajor.disabled = false
                                               niter.disabled = false
                                               cycleniter.disabled = false
                                               threshold.disabled = false
                                               cyclefactor.disabled = false
                                               btns['stop'].disabled = false
                                               slider.disabled = false
                                               image_fig.disabled = false
                                               spectra_fig.disabled = false
                                               if ( ! only_stop ) {
                                                   btns['continue'].disabled = false
                                                   btns['finish'].disabled = false
                                               }
                                           }''',


                       'slider-update': '''if ( '_convergence_data' in convergence_src ) {
                                               // use saved state for update of convergence plot if it is
                                               // available (so update can happen while tclean is running)
                                               update_convergence( )
                                           } else {
                                               // update convergence plot with a request to python
                                               conv_pipe.send( convergence_id,
                                                               { action: 'update', value: [ pos[0], cb_obj.value ] },
                                                                 //      stokes-------------^^^^^^  ^^^^^^^^^^^^^^--------chan
                                                                 update_convergence )
                                           }''',

                       'clean-status-update': '''function update_status( status ) {
                                               const stopstr = [ 'zero stop code',
                                                                 'iteration limit hit',
                                                                 'force stop',
                                                                 'no change in peak residual across two major cycles',
                                                                 'peak residual increased by 3x from last major cycle',
                                                                 'peak residual increased by 3x from the minimum',
                                                                 'zero mask found',
                                                                 'no mask found',
                                                                 'n-sigma or other valid exit criterion',
                                                                 'major cycle limit hit',
                                                                 'unrecognized stop code' ]
                                               if ( typeof status === 'number' ) {
                                                   stopstatus.text = '<div>' +
                                                                     stopstr[ status < 0 || status >= stopstr.length ?
                                                                              stopstr.length - 1 : status ] +
                                                                     '</div>'
                                               } else {
                                                   stopstatus.text = `<div>${status}</div>`
                                               }
                                           }''',

                       'clean-gui-update': '''function update_gui( msg ) {
                                               if ( msg.result === 'no-action' ) {
                                                   update_status( 'nothing done' )
                                                   enable( false )
                                               } else
                                               if ( msg.result === 'update' ) {
                                                   if ( 'cmd' in msg ) {
                                                       log.text = log.text + msg.cmd
                                                   }
                                                   refresh( msg )
                                                   // stopcode == 1: iteration limit hit
                                                   // stopcode == 9: major cycle limit hit
                                                   // *******************************************************************************************
                                                   // ******** perhaps the user should not be locked into exiting after the limit is hit ********
                                                   // *******************************************************************************************
                                                   //state.stopped = state.stopped || (msg.stopcode > 1 && msg.stopcode < 9) || msg.stopcode == 0
                                                   state.stopped = false
                                                   if ( state.mode === 'interactive' && ! state.awaiting_stop ) {
                                                       btns['stop'].button_type = "danger"
                                                       update_status( 'stopcode' in msg ? msg.stopcode : -1 )
                                                       if ( ! state.stopped ) {
                                                           enable( false )
                                                       } else {
                                                           disable( false )
                                                       }
                                                   } else if ( state.mode === 'continuous' && ! state.awaiting_stop ) {
                                                       if ( ! state.stopped ) {
                                                           ctrl_pipe.send( ids[cb_obj.origin.name],
                                                                           { action: 'finish',
                                                                             value: { niter: niter.value, cycleniter: cycleniter.value, nmajor: nmajor.value,
                                                                                      threshold: threshold.value, cyclefactor: cyclefactor.value,
                                                                                      mask: img_src.masks( ),
                                                                                      breadcrumbs: img_src.breadcrumbs( ) } },
                                                                           update_gui )
                                                       } else {
                                                           state.mode = 'interactive'
                                                           btns['stop'].button_type = "danger"
                                                           enable(false)
                                                           state.stopped = false
                                                           update_status( 'stopcode' in msg ? msg.stopcode : -1 )
                                                       }
                                                   }
                                               } else if ( msg.result === 'error' ) {
                                                   img_src.drop_breadcrumb('E')
                                                   if ( 'cmd' in msg ) {
                                                       log.text = log.text + msg.cmd
                                                   }
                                                   state.mode = 'interactive'
                                                   btns['stop'].button_type = "danger"
                                                   state.stopped = false
                                                   update_status( 'stopcode' in msg ? msg.stopcode : -1 )
                                                   enable( false )
                                               }
                                           }''',

                       'clean-wait':    '''function wait_for_tclean_stop( msg ) {
                                               state.mode = 'interactive'
                                               btns['stop'].button_type = "danger"
                                               enable( false )
                                               state.awaiting_stop = false
                                               update_status( 'user requested stop' )
                                           }''',
                   }

        self._cube = CubeMask( self._residual_path, mask=self._mask_path, abort=self._abort_handler )
        ###
        ### error or exception result
        ###
        self._error_result = None

        ###
        ### websocket servers
        ###
        self._control_server = None
        self._converge_server = None

    def _init_pipes( self ):
        if not self.__pipes_initialized:
            self.__pipes_initialized = True
            self._pipe['control'] = DataPipe( address=find_ws_address( ), abort=self._abort_handler )
            self._pipe['converge'] = DataPipe( address=find_ws_address( ), abort=self._abort_handler )

    def _launch_gui( self ):
        '''create and show GUI
        '''
        self._fig = { }

        ###
        ### set up websockets which will be used for control and convergence updates
        ###
        self._init_pipes( )

        ###
        ### Python-side handler for events from the interactive clean control buttons
        ###
        async def clean_handler( msg, self=self ):
            if msg['action'] == 'next' or msg['action'] == 'finish':
                if 'mask' in msg['value']:
                    if 'breadcrumbs' in msg['value'] and msg['value']['breadcrumbs'] != self._last_mask_breadcrumbs:
                        self._last_mask_breadcrumbs = msg['value']['breadcrumbs']
                        mask_dir = "%s.mask" % self._imagename
                        shutil.rmtree(mask_dir)
                        new_mask = self._cube.jsmask_to_raw(msg['value']['mask'])
                        self._mask_history.append(new_mask)

                        msg['value']['mask'] = convert_masks(masks=new_mask, coord='pixel', cdesc=self._cube.coorddesc())

                    else:
                        msg['value']['mask'] = ''
                else:
                    msg['value']['mask'] = ''
                self._clean.update( { **msg['value'] } )
                stopcode, self._convergence_data = await self._clean.__anext__( )
                # *******************************************************************************************
                # ******** perhaps the user should not be locked into exiting after the limit is hit ********
                # *******************************************************************************************
                #if stopcode > 1 and stopcode < 9: # 1: iteration limit hit, 9: major cycle limit hit
                #    self._clean.finalize()

                    # self._cube.update_image(self._clean.finalize()['image']) # TODO show the restored image
                if len(self._convergence_data) == 0 and stopcode == 7:
                    return dict( result='error', stopcode=stopcode, cmd=f"<p>mask error encountered (stopcode {stopcode})</p>", convergence=None  )
                if len(self._convergence_data) == 0:
                    return dict( result='no-action', stopcode= stopcode, cmd=f'<p style="width:790px">no operation</p>',
                                 convergence=None, iterdone=0 )
                if len(self._convergence_data) * len(self._convergence_data[0]) > self._threshold_chan or \
                   len(self._convergence_data[0][0]['iterations']) > self._threshold_iterations:
                    return dict( result='update', stopcode=stopcode, cmd=''.join([ f'<p style="width:790px">{cmd}</p>' for cmd in self._clean.cmds( )[-2:] ]),
                                 convergence=None, iterdone=sum([ x['iterations'][1]  for y in self._convergence_data.values() for x in y.values( ) ]) )
                else:
                    return dict( result='update', stopcode=stopcode, cmd=''.join([ f'<p style="width:790px">{cmd}</p>' for cmd in self._clean.cmds( )[-2:] ]),
                                 convergence=self._convergence_data, iterdone=sum([ x['iterations'][1]  for y in self._convergence_data.values() for x in y.values( ) ]) )

                return dict( result='update', stopcode=stopcode, cmd=''.join([ f'<p style="width:790px">{cmd}</p>' for cmd in self._clean.cmds( )[-2:] ]),
                             convergence=self._convergence_data, iterdone=sum([ x['iterations'][1]  for y in self._convergence_data.values() for x in y.values( ) ]) )
            elif msg['action'] == 'stop':
                self.__stop( )
                return dict( result='stopped', update=dict( ) )
            elif msg['action'] == 'status':
                return dict( result="ok", update=dict( ) )
            else:
                print( "got something else: '%s'" % msg['action'] )

        ###
        ### Setup id that will be used for messages from each button
        ###
        self._ids['clean'] = { }
        for btn in "continue", 'finish', 'stop':
            self._ids['clean'][btn] = str(uuid4( ))
            #print("%s: %s" % ( btn, self._ids['clean'][btn] ) )
            self._pipe['control'].register( self._ids['clean'][btn], clean_handler )

        ###
        ### Retrieve convergence information
        ###
        def convergence_handler( msg, self=self ):
            if msg['value'][1] in self._convergence_data:
                return { 'action': 'update-success',
                         'result': dict(converge=self._convergence_data[msg['value'][1]][msg['value'][0]]) }
                ###                                          chan-------^^^^^^^^^^^^^^^  ^^^^^^^^^^^^^^^-------stokes
            else:
                return { 'action': 'update-failure' }

        self._pipe['converge'].register( self._convergence_id, convergence_handler )

        ###
        ### Data source that will be used for updating the convergence plot
        ###
        self._convergence_source = ColumnDataSource(data=self._convergence_data[0][self._stokes])

        ###
        ### Setup script that will be called when the user closes the
        ### browser tab that is running interactive clean
        ###
        self._pipe['control'].init_script = CustomJS( args=dict( convergence_src=self._convergence_source,
                                                                 ctrl_pipe=self._pipe['control'],
                                                                 ids=self._ids['clean'] ),
                                                      code=self._js['initialize'] )

        self._fig['convergence'] = figure( tooltips=[ ("x","$x"), ("y","$y"), ("value", "(@peakRes, @modelFlux)")],
                                           output_backend="webgl", plot_height=180, plot_width=800,
                                           tools=[ ],
                                           title='Convergence', x_axis_label='Iteration', y_axis_label='Peak Residual' )
        self._fig['convergence'].yaxis.axis_label_text_color = self._color['residual']
        self._fig['convergence'].extra_y_ranges = { 'modelFlux': Range1d( min(self._convergence_source.data['modelFlux'])*0.5,
                                                                     max(self._convergence_source.data['modelFlux'])*1.5 ) }
        self._fig['convergence'].add_layout( LinearAxis( y_range_name='modelFlux', axis_label='Total Flux', axis_label_text_color=self._color['flux'] ), 'right' )

        self._fig['convergence'].circle( x='iterations',
                                         y='peakRes',
                                         color=self._color['residual'],
                                         size=10,
                                         alpha=0.4,
                                         #legend_label='Peak Residual',
                                         source=self._convergence_source )
        self._fig['convergence'].circle( x='iterations',
                                         y='modelFlux',
                                         color=self._color['flux'],
                                         size=10,
                                         alpha=0.4,
                                         y_range_name='modelFlux',
                                         #legend_label='Total Flux',
                                         source=self._convergence_source )
        self._fig['convergence'].line( x='iterations',
                                       y='peakRes',
                                       color=self._color['residual'],
                                       source=self._convergence_source )

        self._fig['convergence'].line( x='iterations',
                                       y='modelFlux',
                                       color=self._color['flux'],
                                       y_range_name='modelFlux',
                                       source=self._convergence_source )

        # TClean Controls
        cwidth = 80
        cheight = 50
        self._control['clean'] = { }
        self._control['clean']['continue'] = Button( label="", max_width=cwidth, max_height=cheight, name='continue',
                                                     icon=SVGIcon(icon_name="iclean-continue", size=2.5) )
        self._control['clean']['finish'] = Button( label="", max_width=cwidth, max_height=cheight, name='finish',
                                                   icon=SVGIcon(icon_name="iclean-finish", size=2.5) )
        self._control['clean']['stop'] = Button( label="", button_type="danger", max_width=cwidth, max_height=cheight, name='stop',
                                                 icon=SVGIcon(icon_name="iclean-stop", size=2.5) )
        width = 35
        height = 35
        self._control['help'] = Button( label="", max_width=width, max_height=height, name='help',
                                        icon=SVGIcon(icon_name='help', size=1.4) )


        self._control['nmajor'] = TextInput( title='nmajor', value="%s" % self._params['nmajor'], width=90 )
        self._control['niter'] = TextInput( title='niter', value="%s" % self._params['niter'], width=90 )
        self._control['cycleniter'] = TextInput( title="cycleniter", value="%s" % self._params['cycleniter'], width=90 )
        self._control['threshold'] = TextInput( title="threshold", value="%s" % self._params['threshold'], width=90 )
        self._control['cycle_factor'] = TextInput( value="%s" % self._params['cyclefactor'], title="cyclefactor", width=90 )

        self._control['goto'] = TextInput( title="goto channel", value="", width=90 )

        self._fig['slider'] = self._cube.slider( )
        self._fig['image'] = self._cube.image( )
        self._fig['image-source'] = self._cube.js_obj( )
        self._fig['spectra'] = self._cube.spectra( )

        self._cb['clean'] = CustomJS( args=dict( btns=self._control['clean'],
                                                 state=dict( mode='interactive', stopped=False, awaiting_stop=False, mask="" ),
                                                 ctrl_pipe=self._pipe['control'], conv_pipe=self._pipe['converge'],
                                                 ids=self._ids['clean'],
                                                 img_src=self._fig['image-source'],
                                                 #spec_src=self._image_spectra,
                                                 niter=self._control['niter'], cycleniter=self._control['cycleniter'],
                                                 nmajor=self._control['nmajor'],
                                                 threshold=self._control['threshold'], cyclefactor=self._control['cycle_factor'],
                                                 convergence_src=self._convergence_source, convergence_id=self._convergence_id,
                                                 convergence_fig=self._fig['convergence'],
                                                 log=self._status['log'],
                                                 slider=self._fig['slider'],
                                                 image_fig=self._fig['image'],
                                                 spectra_fig=self._fig['spectra'],
                                                 stopstatus=self._status['stopcode'],
                                                 cube_obj = self._cube.js_obj( )
                                                ),
                                      code=self._js['update-converge'] + self._js['clean-refresh'] + self._js['clean-disable'] +
                                           self._js['clean-enable'] + self._js['clean-status-update'] +
                                           self._js['clean-gui-update'] + self._js['clean-wait'] +
                                           '''if ( ! state.stopped && cb_obj.origin.name == 'finish' ) {
                                                  state.mode = 'continuous'
                                                  update_status( 'running multiple iterations' )
                                                  disable( false )
                                                  btns['stop'].button_type = "warning"
                                                  ctrl_pipe.send( ids[cb_obj.origin.name],
                                                                  { action: 'finish',
                                                                    value: { niter: niter.value, cycleniter: cycleniter.value, nmajor: nmajor.value,
                                                                             threshold: threshold.value, cyclefactor: cyclefactor.value,
                                                                             mask: img_src.masks( ),
                                                                             breadcrumbs: img_src.breadcrumbs( ) } },
                                                                  update_gui )
                                              }
                                              if ( ! state.stopped && state.mode === 'interactive' &&
                                                   cb_obj.origin.name === 'continue' ) {
                                                  update_status( 'running one iteration' )
                                                  disable( true )
                                                  // only send message for button that was pressed
                                                  // it's unclear whether 'this.origin.' or 'cb_obj.origin.' should be used
                                                  // (or even if 'XXX.origin.' is public)...
                                                  ctrl_pipe.send( ids[cb_obj.origin.name],
                                                                  { action: 'next',
                                                                    value: { niter: niter.value, cycleniter: cycleniter.value, nmajor: nmajor.value,
                                                                             threshold: threshold.value, cyclefactor: cyclefactor.value,
                                                                             mask: img_src.masks( ),
                                                                             breadcrumbs: img_src.breadcrumbs( ) } },
                                                                  update_gui )
                                              }
                                              if ( state.mode === 'interactive' && cb_obj.origin.name === 'stop' ) {
                                                  disable( true )
                                                  //ctrl_pipe.send( ids[cb_obj.origin.name],
                                                  //                { action: 'stop',
                                                  //                  value: { } },
                                                  //                update_gui )
                                                  img_src.done( )
                                              } else if ( state.mode === 'continuous' &&
                                                          cb_obj.origin.name === 'stop' &&
                                                          ! state.awaiting_stop ) {
                                                  disable( true )
                                                  state.awaiting_stop = true
                                                  ctrl_pipe.send( ids[cb_obj.origin.name],
                                                                  { action: 'status',
                                                                    value: { } },
                                                                  wait_for_tclean_stop )
                                              }''' )

        self._control['clean']['continue'].js_on_click( self._cb['clean'] )
        self._control['clean']['finish'].js_on_click( self._cb['clean'] )
        self._control['clean']['stop'].js_on_click( self._cb['clean'] )

        self._fig['slider'].js_on_change( 'value',
                                          CustomJS( args=dict( convergence_src=self._convergence_source,
                                                               convergence_fig=self._fig['convergence'],
                                                               conv_pipe=self._pipe['converge'], convergence_id=self._convergence_id,
                                                               img_src=self._fig['image-source']
                                                             ),
                                                    code='''const pos = img_src.cur_chan;''' +             ### later we will receive the polarity from some widget mechanism...
                                                            self._js['update-converge'] + self._js['slider-update'] ) )
        
        # Generates the HTML for the controls layout:
        # nmajor niter cycleniter cycle_factor threshold  -----------
        #                  slider                         -  image  -    help
        # goto    continue finish stop                    -----------    help
        # status: stopcode                                               help
        # stats                                                          help
        # spectra
        # convergence
        # log

        mask_color_pick, mask_alpha_pick, mask_clean_notclean_pick = self._cube.bitmask_controls( )

        self._fig['layout'] = column(
                                  row(
                                      column(
                                          row( self._control['nmajor'],
                                               self._control['niter'],
                                               Spacer(width=10, sizing_mode='scale_height'),
                                               self._control['clean']['continue'],
                                               self._control['clean']['finish'] ),
                                          row( self._control['cycleniter'],
                                               self._control['cycle_factor'],
                                               self._control['threshold'],
                                               self._control['clean']['stop'] ),
                                          row( self._fig['slider'], self._control['goto'] ),
                                          row ( Div( text="<div><b>status:</b></div>" ), self._status['stopcode'] ),
                                          self._cube.statistics( sizing_mode = "stretch_height" ),
                                      ),
                                      column( self._fig['image'],
                                              row( Spacer(height=self._control['help'].height, sizing_mode="scale_width"),
                                                   self._cube.palette( ),
                                                   mask_clean_notclean_pick,
                                                   mask_color_pick,
                                                   mask_alpha_pick,
                                                   self._control['help'],
                                                   Spacer(height=self._control['help'].height, width=30)
                                              )
                                      ),
                                      self._cube.help( width=450, height=100,
                                                       rows=[ '<tr><td><i><b>red</b> stop button</i></td><td>clicking the stop button (when red) will close the dialog and control to python</td></tr>',
                                                              '<tr><td><i><b>orange</b> stop button</i></td><td>clicking the stop button (when orang) will return control to the GUI after the currently executing tclean run completes</td></tr>',

                                                             ]
                                                     )
                                  ),
                                  self._fig['spectra'],
                                  self._fig['convergence'],
                                  Spacer(width=380, height=40, sizing_mode='scale_width'),
                                  self._status['log'] )

        self._control['help'].js_on_click( CustomJS( args=dict( help=self._cube.help( ) ),
                                                     code='''if ( help.visible == true ) help.visible = false
                                                             else help.visible = true''' ) )

        self._control['goto'].js_on_change( 'value', CustomJS( args=dict( img=self._cube.js_obj( ),
                                                                          slider=self._fig['slider'],
                                                                          status=self._status['stopcode'] ),
                                                               code='''let values = cb_obj.value.split(/[ ,]+/).map((v,) => parseInt(v))
                                                                       if ( values.length > 2 ) {
                                                                         status._error_set = true
                                                                         status.text = '<div>enter at most two indexes</div>'
                                                                       } else if ( values.filter((x) => x < 0 || isNaN(x)).length > 0 ) {
                                                                         status._error_set = true
                                                                         status.text = '<div>invalid channel entered</div>'
                                                                       } else {
                                                                         if ( status._error_set ) {
                                                                           status._error_set = false
                                                                           status.text = '<div/>'
                                                                         }
                                                                         if ( values.length == 1 ) {
                                                                           if ( values[0] >= 0 && values[0] < img.num_chans[1] ) {
                                                                             status.text= `<div>moving to channel ${values[0]}</div>`
                                                                             slider.value = values[0]
                                                                           } else {
                                                                             status._error_set = true
                                                                             status.text = `<div>channel ${values[0]} out of range</div>`
                                                                           }
                                                                         } else if ( values.length == 2 ) {
                                                                           if ( values[0] < 0 || values[0] >= img.num_chans[1] ) {
                                                                             status._error_set = true
                                                                             status.text = `<div>channel ${values[0]} out of range</div>`
                                                                           } else {
                                                                             if ( values[1] < 0 || values[1] >= img.num_chans[0] ) {
                                                                               status._error_set = true
                                                                               status.text = `<div>stokes ${values[1]} out of range</div>`
                                                                             } else {
                                                                               status.text= `<div>moving to channel ${values[0]}/${values[1]}</div>`
                                                                               slider.value = values[0]
                                                                               img.channel( values[0], values[1] )
                                                                             }
                                                                           }
                                                                         }
                                                                       }''' ) )
        self._cube.connect( )
        show(self._fig['layout'])

    def _asyncio_loop( self ):
        async def async_loop( f1, f2, f3 ):
            return await asyncio.gather( f1, f2, f3 )

        self._control_server = websockets.serve( self._pipe['control'].process_messages, self._pipe['control'].address[0], self._pipe['control'].address[1] )
        self._converge_server = websockets.serve( self._pipe['converge'].process_messages, self._pipe['converge'].address[0], self._pipe['converge'].address[1] )
        resource_manager.reg_webserver(self._control_server.ws_server)
        resource_manager.reg_webserver(self._converge_server.ws_server)
        return async_loop( self._control_server, self._converge_server, self._cube.loop( ) )

    def __call__( self, loop=asyncio.get_event_loop( ) ):
        '''Display GUI using the event loop specified by ``loop``.

        Parameters
        ----------
        loop: event loop object
            defaults to standard asyncio event loop.

        Example:
            Create ``iclean`` object and display::

                print( "Result: %s" %
                       iclean( vis='refim_point_withline.ms', imagename='test', imsize=512,
                               cell='12.0arcsec', specmode='cube',
                               interpolation='nearest', ... )( ) )
        '''
        try:
            loop.run_until_complete(self.show( ))
            loop.run_forever( )
        except KeyboardInterrupt:
            print('\nInterrupt received, stopping GUI...')

        return self.result( )

    def show( self ):
        '''Get the InteractiveClean event loop to use for running the interactive clean GUI
        as part of an external event loop.
        '''
        self._launch_gui( )
        return self._asyncio_loop( )

    def result( self ):
        '''If InteractiveClean had a return value, it would be filled in as part of the
        GUI dialog between Python and JavaScript and this function would return it'''
        if isinstance(self._error_result,Exception):
            raise self._error_result
        elif self._error_result is not None:
            return self._error_result
        return self._cube.result( )

    def masks( self ):
        '''Retrieves the masks which were used with interactive clean.

        Returns
        -------
        The standard ``casagui`` cube region dictionary which contains two elements
        ``masks`` and ``polys``.

        The value of the ``masks`` element is a dictionary that is indexed by
        tuples of ``(stokes,chan)`` and the value of each element is a list
        whose elements describe the polygons drawn on the channel represented
        by ``(stokes,chan)``. Each polygon description in this list has a
        polygon index (``p``) and a x/y translation (``d``).

        The value of the ``polys`` element is a dictionary that is indexed by
        polygon indexes. The value of each polygon index is a dictionary containing
        ``type`` (whose value is either ``'rect'`` or ``'poly``) and ``geometry``
        (whose value is a dictionary containing ``'xs'`` and ``'ys'`` (which are
        the x and y coordinates that define the polygon).

        This can be converted to other formats with ``casagui.utils.convert_masks``.
        '''
        return copy.deepcopy(self._mask_history)    ## don't allow users to change history

    def history( self ):
        '''Retrieves the commands used during the interactive clean session.

        Returns
        -------
        list[str]  tclean calls made during the interactive clean session.
        '''
        return self._clean.cmds( )



########################################################################
#
# Copyright (C) 2022
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: aips2-request@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''Functions for palette access.'''
from bokeh import palettes as __palettes
from ...utils import static_vars

__palette_list = [ p for p in dir(__palettes) if isinstance(getattr(__palettes,p), tuple) ]

def available_palettes ( ):
    return __palette_list

def find_palette( name ):
    if name not in __palette_list:
        return None
    return getattr( __palettes, name )

@static_vars(result=None)
def default_palette( ):
    if default_palette.result is None:
        result = None         ### look for palette called 'Plasma...' with the most colors
        fallback = None       ### if none is found fall back to the longest palette name
        for p in __palette_list:
            if fallback is None:
                fallback = p
            elif len(p) > len(fallback):
                fallback = p
            elif len(p) == len(fallback) and p > fallback:
                fallback = p
            if result is None:
                if p.startswith('Plasma'):
                    result = p
            elif p.startswith('Plasma'):
                if len(p) > len(result):
                    result = p
                elif len(p) == len(result) and p > result:
                    result = p
            if result is not None:
                default_palette.result = result
            else:
                default_palette.result = fallback
    return default_palette.result



from bokeh.core.properties import Tuple, String, Int
from bokeh.layouts import column
from bokeh.models import Button, CustomJS
from bokeh.io import show

class ICleanButton(Button):
    """
    Custom extension of Bokeh Button that implements a websocket connection. For information on 
    Bokeh button see,
        https://docs.bokeh.org/en/latest/docs/reference/models/widgets.buttons.html

    Args:
        Button (Class): Basic button class from Bokeh.
    """
     

    __implementation__ = "iclean_button.ts"

    socket = Tuple(String, Int, help="""
    A tuple describing the websocket connection info: (string: address, port: int)
    """)


###
### This is from the "First Look at Imaging" in the casaguides:
###
###    * https://casaguides.nrao.edu/index.php/First_Look_at_Imaging
###
import os
import asyncio
import urllib
import tarfile

from casagui import iclean

##
## demo measurement set to use, from:
##
##     https://casaguides.nrao.edu/index.php?title=M100_Band3_Combine_6.1
##
## $ wget https://bulk.cv.nrao.edu/almadata/sciver/M100Band3_12m/M100_Band3_12m_CalibratedData.tgz
## $ wget https://bulk.cv.nrao.edu/almadata/sciver/M100Band3ACA/M100_Band3_7m_CalibratedData.tgz
## $ tar -xf M100_Band3_12m_CalibratedData.tgz
## $ tar -xf M100_Band3_7m_CalibratedData.tgz
## $ ln -s M100_Band3_7m_CalibratedData/M100_Band3_7m_CalibratedData.ms/
## $ ln -s M100_Band3_12m_CalibratedData/M100_Band3_12m_CalibratedData.ms/
## $ casa -c "split(vis='M100_Band3_12m_CalibratedData.ms', outputvis='M100_12m_CO.ms',spw='0',field='M100', datacolumn='data',keepflags=False); split(vis='M100_Band3_7m_CalibratedData.ms', outputvis='M100_7m_CO.ms',spw='3,5',field='M100', datacolumn='data',keepflags=False); concat(vis=['M100_12m_CO.ms','M100_7m_CO.ms'], concatvis='M100_combine_CO.ms')"
##
ms_path = 'M100_combine_CO.ms'
##
## where to fetch the demo measurement set
##
##
## Using the URL given in the casaguide results in the error:
##
##      File "first_look_interactive.py", line 28, in <module>
##        tar = tarfile.open(fileobj=tstream, mode="r")
##      File "/opt/local/Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/tarfile.py", line 1601, in open
##        saved_pos = fileobj.tell()
##      io.UnsupportedOperation: seek

#ms_url = "https://bulk.cv.nrao.edu/almadata/public/working/sis14_twhya_calibrated_flagged.ms.tar"
#ms_url = "https://casa.nrao.edu/download/devel/casagui/sis14_twhya_calibrated_flagged.ms.tar.gz"
ms_url = None
##
## output image file name
##
img = 'test'

if not os.path.isdir(ms_path):
    tstream = urllib.request.urlopen(ms_url)
    tar = tarfile.open(fileobj=tstream, mode="r:gz")
    tar.extractall()


ic = iclean( vis=ms_path, imagename=img,
             imsize=800,
             cell=['0.1arcsec'],
             specmode='cube',
             restfreq='115.271201800GHz',
             interpolation='nearest', 
             pblimit=-1e-05,
             gridder='mosaic',
             deconvolver='hogbom',
             nchan=70,
             niter=50,
             cycleniter=10,
             cyclefactor=3,
             threshold='0.0mJy' )

if True:
    print( "Result: %s" % ic( ) )
else:
    try:
        asyncio.get_event_loop().run_until_complete(ic.show( ))
        asyncio.get_event_loop().run_forever()

    except KeyboardInterrupt:
        print('\nInterrupt received, shutting down ...')
        #os.system('rm -rf {output_image}.* *.html *.log'.format(output_image=output_image))

    print( "Result: %s" % ic.result( ) )




from itertools import product
from matplotlib.path import Path
from math import floor,ceil

def polygon_indexes( xs, ys, shape ):
    '''
    Returns indexes for a 2D array of the given ``shape`` which
    lie within the polygon specified by ``xs`` and ``ys``.

    Parameters
    ----------
    xs: list of numbers
        the X coordinates for the vertices of the polygon
    ys: list of numbers
        the Y coordinates for the vertices of the polygon
    shape: ( int, int )
        the shape of the plane in which the polygon is found

    Returns
    -------
    generator of tuples:
        the stream of tuples that is returned will be the indexes
        of the elements which lie within the polygon
    '''
    assert len(shape) == 2, 'contains only works for 2D shapes, so "shape" should have length equal to two'
    assert len(xs) == len(ys), 'to specify a polygon the number of X values must equal the number of Y values'
    path = Path(list(zip(xs,ys)))
    xmin, xmax = min(xs), max(xs)
    ymin, ymax = min(ys), max(ys)
    return filter( path.contains_point, product(range(floor(xmin),ceil(xmax)), range(floor(ymin),ceil(ymax))) )



"""
part of plotants module
"""

from bokeh.models import ColumnDataSource
from bokeh.models.expressions import PolarTransform


def plotants_logpos():
    """
    Makes a logpos plot for plotants
    Returns:

    """
    # code from pipeline summary.py
    # PlotAntsChart draw_polarlog_ant_map_in_subplot
    if "VLA" in telescope:
        # For (E)VLA, set a fixed local center position that has been
        # tuned to work well for its array configurations (CAS-7479).
        xcenter, ycenter = -32, 0
        rmin_min, rmin_max = 12.5, 350
    else:
        # For non-(E)VLA, take the median of antenna offsets as the
        # center for the plot.
        xcenter = np.median(xpos)
        ycenter = np.median(ypos)
        rmin_min, rmin_max = 3, 350

    # Derive radial offset w.r.t. center position.
    r = ((xpos - xcenter) ** 2 + (ypos - ycenter) ** 2) ** 0.5
    # Set rmin, clamp between a min and max value, ignore station
    # at r=0 if one is there.
    rmin = min(rmin_max, max(rmin_min, 0.8 * np.min(r[r > 0])))
    # Update r to move any points below rmin to r=rmin.
    r[r <= rmin] = rmin
    rmin = np.log(rmin)
    # Set rmax.
    rmax = np.log(1.5 * np.max(r))
    # Derive angle of offset w.r.t. center position.
    theta = np.arctan2(xpos - xcenter, ypos - ycenter)
    source = ColumnDataSource(data=dict(r=[], theta=[], labels=[]))
    source.data = dict(r=r, theta=theta, labels=names)
    # https://github.com/bokeh/bokeh/issues/657
    # https://stackoverflow.com/questions/56343933/bokeh-second-legend-showing-scatter-radius
    plot = figure(plot_height=400, plot_width=400, x_axis_type=None, y_axis_type=None)
    # polarx = CustomJSTransform(
    #     args=dict(source=source),
    #     v_func="""
    #     const new_xs = new Array(source.data.r.length)
    #     for(var i = 0; i < new_xs.length; i++) {
    #         new_xs[i] = source.data.r[i] * Math.sin(source.data.theta[i] )
    #     }
    #     return new_xs
    #     """)
    # polary = CustomJSTransform(
    #     args=dict(source=source),
    #     v_func="""
    #     const new_ys = new Array(source.data.r.length)
    #     for(var i = 0; i < new_ys.length; i++) {
    #         new_ys[i] = source.data.r[i] * Math.cos(source.data.theta[i] )
    #     }
    #     return new_ys
    #     """)
    # plot.scatter(
    #     x=transform("r", polarx),
    #     y=transform("r", polary),
    #     source=source,
    #     size=5,
    #     line_color="red",
    #     fill_color="red",
    #     fill_alpha=0.5,
    # )
    # # draw circles
    # show_circle = True
    # circles = [1e5, 3e5, 1e6, 3e6, 1e7] if telescope == "VLBA" else [30, 100, 300, 10000]
    # circles_inner_list = []
    # circles_outer_list = []
    # for circle in circles:
    #     if circle > np.min(r) and show_circle:
    #         if circle < 1000:
    #             circles_inner_list.append(circle)
    #         else:
    #             circles_outer_list.append(circle)
    #     if np.log(circle) > rmax:
    #         show_circle = False
    #
    # plot.circle(0, 0, radius=circles_inner_list, fill_color=None, line_color="black")
    # plot.circle(0, 0, radius=circles_outer_list, fill_color=None, line_color="black")
    # circles_inner_list_labels = [str(r) + "m" for r in circles_inner_list[:-1]]
    # circles_outer_list_labels = [str(r) + "m" for r in circles_outer_list[:-1]]
    # plot.text(0, circles_inner_list[:-1], circles_inner_list_labels, text_font_size="11px", text_align="center", text_baseline="middle")
    # plot.text(0, circles_inner_list[:-1], circles_outer_list_labels, text_font_size="11px", text_align="center", text_baseline="middle")

    t = PolarTransform()
    plot.scatter(
        x=t.x,
        y=t.y,
        source=source,
        size=5,
        line_color="red",
        fill_color="red",
        fill_alpha=0.5,
    )

    # labels = LabelSet(x=transform("r", polarx), y=transform("r", polary), text="labels", x_offset=5, y_offset=5, source=source)
    # labels = LabelSet(x=t.x, y=t.y, text="labels", x_offset=5, y_offset=5, source=source)
    plot.title.text = telescope
    # plot.add_layout(labels)
    return plot



########################################################################
#
# Copyright (C) 2021,2022
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: aips2-request@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''Implementation of ``ImagePipe`` class which provides a ``websockets``
implementation for CASA images which allows for interacitve display
of image cube channels in response to user input.'''

import json
import asyncio

from bokeh.models.sources import DataSource
from bokeh.util.compiler import TypeScript
from bokeh.core.properties import Tuple, String, Int, Instance, Nullable
from bokeh.models.callbacks import Callback

import numpy as np
try:
    import casatools as ct
    from casatools import regionmanager
    from casatools import image as imagetool
except:
    ct = None
    from casagui.utils import warn_import
    warn_import('casatools')

from ..utils import pack_arrays
from ...utils import partition, resource_manager

class ImagePipe(DataSource):
    """The `ImagePipe` allows for updates to Bokeh plots from a CASA or CNGI

    image. This is done using a `websocket`. A `ImagePipe` is created with
    the path to the image, and then it is used as the input to an
    `ImageDataSource` or a `SpectraDataSource`. This allows a single CASA
    or CNGI imge to be opened once and shared among multiple Bokeh plots,
    for example ploting an image channel and a plot of a spectra from the
    image cube.

    Attributes
    ----------
    address: tuple of string and int
        the string is the IP address for the network that should be used and the
        integer is the port number, see ``casagui.utils.find_ws_address``
    init_script: JavaScript
        this javascript is run when this DataPipe object is initialized. init_script
        is used to run caller JavaScript which needs to be run at initialization time.
        This is optional and does not need to be set.
    """
    __im_path = None
    __im = None
    __chan_shape = None

    init_script = Nullable(Instance(Callback), help="""
    JavaScript to be run during initialization of an instance of an DataPipe object.
    """)

    address = Tuple( String, Int, help="two integer sequence representing the address and port to use for the websocket" )
    shape = Tuple( Int, Int, Int, Int, help="shape: [ RA, DEC, Stokes, Spectral ]" )

    __implementation__ = TypeScript( "" )

    def __open_image( self, image ):
        if self.__img is not None:
            self.__img.close( )
        self.__img = imagetool( )
        self.__rgn = regionmanager( )
        try:
            self.__img.open(image)
            self.__image_path = image
        except Exception as ex:
            self.__img = None
            self.__image_path = None
            raise RuntimeError(f'could not open image: {image}') from ex
        imshape = self.__img.shape( )
        if self.__msk is not None and all(self.__msk.shape( ) != imshape):
            raise RuntimeError(f'mismatch between image shape ({imshape}) and mask shape ({self.__msk.shape( )})')
        if self.__chan_shape is None: self.__chan_shape = list(imshape[0:2])

    def __open_mask( self, mask ):
        if mask is None:
            self.__mask_path = None
            return
        if self.__msk is not None:
            self.__msk.close( )
        self.__msk = imagetool( )
        try:
            self.__msk.open(mask)
            self.__mask_path = mask
        except Exception as ex:
            self.__msk = None
            self.__mask_path = None
            raise RuntimeError(f'could not open mask: {mask}') from ex
        mskshape = self.__msk.shape( )
        if self.__img is not None and all(self.__img.shape( ) != mskshape):
            raise RuntimeError(f'mismatch between image shape ({self.__img.shape( )}) and mask shape ({mskshape})')
        if self.__chan_shape is None: self.__chan_shape = list(mskshape[0:2])

    def channel( self, index ):
        """Retrieve one channel from the image cube. The `index` should be a
        two element list of integers. The first integer is the ''stokes'' axis
        in the image cube. The second integer is the ''channel'' axis in the
        image cube.

        Parameters
        ----------
        index: [ int, int ]
            list containing first the ''stokes'' index and second the ''channel'' index
        """
        if self.__img is None:
            raise RuntimeError('no image is available')
        return np.squeeze( self.__img.getchunk( blc=[0,0] + index,
                                                trc=self.__chan_shape + index) ).transpose( )

    def mask( self, index, modify=False ):
        """Retrieve one channel mask from the mask cube. The `index` should be a
        two element list of integers. The first integer is the ''stokes'' axis
        in the image cube. The second integer is the ''channel'' axis in the
        image cube.

        Parameters
        ----------
        index: [ int, int ]
            list containing first the ''stokes'' index and second the ''channel'' index
        modify: boolean
            If true, it implies that the channel mask is being retrieved for modification
            and updating the channel on disk. If false, it implies that the channel mask
            is being retrieved for display.
        """
        if self.__msk is None:
            raise RuntimeError(f'cannot retrieve mask at {repr(index)} because no mask cube exists')
        return ( np.squeeze( self.__msk.getchunk( blc=[0,0] + index,
                                                trc=self.__chan_shape + index) ).astype(np.bool_).transpose( )
                 if modify == False else
                 np.squeeze( self.__msk.getchunk( blc=[0,0] + index,
                                                trc=self.__chan_shape + index) ) )

    def put_mask( self, index, mask ):
        """Replace one channel mask with the mask specified as the second parameter.
        The `index` should be a two element list of integers. The first integer is the
        ''stokes'' axis in the image cube. The second integer is the ''channel'' axis
        in the image cube. The assumption is that the :code:`mask` parameter was retrieved
        from the mask cube using the :code:`mask(...)` function with the :code:`modify`
        parameter set to :code:`True`.

        Parameters
        ----------
        index: [ int, int ]
            list containing first the ''stokes'' index and second the ''channel'' index
        mask: numpy.ndarray
            two dimensional array to replace the existing mask for the channel specified
            by :code:`index`
        """
        if self.__msk is None:
            raise RuntimeError(f'cannot replace mask at {repr(index)} because no mask cube exists')
        self.__msk.putchunk( blc=[0,0] + index, pixels=mask )

    def spectra( self, index ):
        """Retrieve one spectra from the image cube. The `index` should be a
        three element list of integers. The first integer is the ''right
        ascension'' axis, the second integer is the ''declination'' axis,
        and the third integer is the ''stokes'' axis.

        Parameters
        ----------
        index: [ int, int, int ]
            list containing first the ''right ascension'', the ''declination'' and
            the ''stokes'' axis
        """
        index = list(map( lambda i: 0 if i is None else i, index ))
        if index[0] >= self.shape[0]:
            index[0] = self.shape[0] - 1
        if index[1] >= self.shape[1]:
            index[1] = self.shape[1] - 1
        if self.__img is None:
            raise RuntimeError('no image is available')
        result = np.squeeze( self.__img.getchunk( blc=index + [0],
                                                 trc=index + [self.shape[-1]] ) )
        ### should return spectral freq etc.
        ### here for X rather than just the index
        try:
            return { 'x': range(len(result)), 'y': list(result) }
        except Exception:
            ## In this case, result is not iterable (e.g.) only one channel in the cube.
            ## A zero length numpy ndarray has no shape and looks like a float but it is
            ## an ndarray.
            return { 'x': [0], 'y': [float(result)] }

    def __init__( self, image, *args, mask=None, abort=None, stats=False, **kwargs ):
        super( ).__init__( *args, **kwargs, )

        if ct is None:
            raise RuntimeError('cannot open an image because casatools is not available')

        self.__img = None
        self.__msk = None
        resource_manager.reg_at_exit(self, '__del__')
        self._stats = stats
        self.__open_image( image )
        self.__open_mask( mask )
        self.shape = list(self.__img.shape( ))
        self.__session = None
        self.__abort = abort

        if self.__abort is not None and not callable(self.__abort):
            raise RuntimeError('abort function must be callable')

    def __del__(self):
        if self.__rgn:
            self.__rgn.done( )
        if self.__img != None:
            self.__img.close()
            self.__img.done()
            self.__img = None

    def coorddesc( self ):
        ia = imagetool( )
        ia.open(self.__image_path)
        csys = ia.coordsys( )
        ia.close( )
        return { 'csys': csys, 'shape': tuple(self.shape) }

    def statistics( self, index ):
        """Retrieve statistics for one channel from the image cube. The `index`
        should be a two element list of integers. The first integer is the
        ''stokes'' axis in the image cube. The second integer is the ''channel''
        axis in the image cube.

        Parameters
        ----------
        index: [ int, int ]
            list containing first the ''stokes'' index and second the ''channel'' index
        """
        def singleton( potential_nonlist ):
            # convert a list of a single element to the element
            return potential_nonlist if len(potential_nonlist) != 1 else potential_nonlist[0]
        def sort_result( unsorted_dictionary ):
            part = partition( lambda s: (s.startswith('trc') or s.startswith('blc')), sorted(unsorted_dictionary.keys( )) )
            return { k: unsorted_dictionary[k] for k in part[1] + part[0] }

        reg = self.__rgn.box( [0,0] + index, self.__chan_shape + index )
        ###
        ### This seems like it should work:
        ###
        #      rawstats = self.__img.statistics( region=reg )
        ###
        ### but it does not so we have to create a one-use image tool (see CAS-13625)
        ###
        ia = imagetool( )
        ia.open(self.__image_path)
        rawstats = ia.statistics( region=reg )
        ia.close( )
        return sort_result( { k: singleton([ x.item( ) for x in v ]) if isinstance(v,np.ndarray) else v for k,v in rawstats.items( ) } )

    async def process_messages( self, websocket ):
        """Process messages related to image display updates.

        Parameters
        ----------
        websocket: websocket object
            Websocket serve function passes in the websocket object to use.
        path:
            Websocket serve provides a second parameter.
        """
        count = 1
        async for message in websocket:
            cmd = json.loads(message)
            if 'session' not in cmd:
                await websocket.close( )
                err = RuntimeError(f'session not in: {cmd}')
                if self.__abort is not None:
                    self.__abort( asyncio.get_running_loop( ), err )
                else:
                    raise err
                return
            elif self.__session != None and self.__session != cmd['session']:
                await websocket.close( )
                err = RuntimeError(f"session corruption: {cmd['session']} does not equal {self.__session}")
                if self.__abort is not None:
                    self.__abort( asyncio.get_running_loop( ), err )
                else:
                    raise err
                return

            if cmd['action'] == 'channel':
                chan = self.channel(cmd['index'])
                mask = self.mask(cmd['index'])
                if self._stats:
                    #statistics for the displayed plane of the image cubea
                    statistics = self.statistics( cmd['index'] )
                    msg = { 'id': cmd['id'],
                            # 'stats': pack_arrays(self.__img.statistics( axes=cmd['index'] )),
                            'message': { 'chan': { 'img': [ pack_arrays(chan) ], 'msk': [ pack_arrays(mask) ] },
                                         'stats': { 'labels': list(statistics.keys( )), 'values': pack_arrays(list(statistics.values( ))) } } }
                else:
                    msg = { 'id': cmd['id'],
                            'message': { 'chan': { 'img': [ pack_arrays(chan) ], 'msk': [ pack_arrays(mask) ] } } }

                await websocket.send(json.dumps(msg))
                count += 1
            elif cmd['action'] == 'spectra':
                msg = { 'id': cmd['id'],
                        'message': { 'spectrum': pack_arrays( self.spectra(cmd['index']) ) } }
                await websocket.send(json.dumps(msg))
            elif cmd['action'] == 'initialize':
                ###
                ### initialize session identifier
                ###
                if self.__session == None:
                    self.__session = cmd['session']
            else:
                print(f"received messate in python with unknown 'action' value: {cmd}")



from bokeh.core.properties import Tuple, String, Int
from bokeh.layouts import column
from bokeh.models import Slider
from bokeh.io import show

class ICleanSlider(Slider):
    """
    Custom extension of Bokeh Slider that implements a websocket connection. For information on 
    Bokeh slider see,
        https://docs.bokeh.org/en/latest/docs/reference/models/widgets.sliders.html

    Args:
        Slider (Class): Basic numerical slider class from Bokeh.
    """
     

    __implementation__ = "iclean_slider.ts"

    socket = Tuple(String, Int, help="""
    A tuple describing the websocket connection info: (string: address, port: int)
    """)